
/* SETI@home plugin for GKrellM
 |
 |  Copyright (c) 2000 Xavier Serpaggi
 |
 |  Author: Xavier Serpaggi <xavier.serpaggi@libertysurf.fr>
 |
 |  Version: 0.6.2
 |
 |  This program is free software which I release under the GNU General Public
 |  License. You may redistribute and/or modify this program under the terms
 |  of that license as published by the Free Software Foundation; either
 |  version 2 of the License, or (at your option) any later version.
 |
 |  This program is distributed in the hope that it will be useful,
 |  but WITHOUT ANY WARRANTY; without even the implied warranty of
 |  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 |  GNU General Public License for more details.
 |  
 |  You should have received a copy of the GNU General Public License
 |  along with this program; if not, write to the Free Software
 |  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
 
/* seti.c       - requires GKrellM 0.10.0 or later.
 |  If this file is compiled as a shared object file and installed in
 |  in your ~/.gkrellm/plugins directory, then it will be loaded when
 |  gkrellm starts up.
 |     gcc -fPIC `gtk-config --cflags` `imlib-config --cflags-gdk` -c seti.c
 |     gcc -shared -Wl -o seti.so seti.o
 |     cp seti.so ~/.gkrellm/plugins
 */


#include "seti.h"

static struct lconv	*locale_conv ;

static GtkWidget	*seti_path_entry ;
static GtkWidget	*seti_args_entry ;

/* avflinsch@att.net -- fix for client path */
static GtkWidget	*seti_client_entry;


static GtkTooltips	*tooltip ;
static gchar		*tooltip_text ;


GdkImlibImage		*krell_image ;
GdkImlibImage		*bg_meter_image ;


static Panel		*panel;
static gint		 style_id;


static gchar		*seti_file_path;
static gchar		*state_file;
static gchar		*user_file;
static gchar		*seti_args;

/* avflinsch@att.net -- fix for client path */
static gchar		*seti_client_path;

static gchar		 seti_first_create = 1 ;
static gchar		*seti_info_text[] =
{
"<b>SETI@home progress meter:\n\n",
"SETI@home progress meter shows the current computation progress of\n",
"a seti-workunit.\n\n",
"Current workunit and completion percent is available via ",
"a tooltip\n\n",
"Progress value is read every minute.\n\n",
"User can change paths to the SETI@home client and state.sah file\n",
"using the ",
"<b>Configs",
" tab.\n\n",
"SETI@home client can be started and stopped (killed) by pressing\n",
"left mouse button over the SETI@home plugin monitor.\n\n",
"Arguments passed to the client when started may be changed in\n",
"the ",
"<b>Configs",
" tab.\n\n",
"<b>Theme:\n\n",
"You can theme you SETI@home progress meter by simply creating\n",
"a ",
"<i>seti",
" directory under your favorite theme directory.\n\n",
"There, you have to put 2 files:\n  -",
"<i>bg_meter.ext:",
" which is the background of the meter,\n  -",
"<i>krell.ext:",
" which will be used to replace the flying saucer in this meter.\n\n",
"Where ",
"<i>.ext",
" represents a GKrellM understandable file format like png or xpm \n\n",
"You can also change the properties of the label by adding lines to\n",
"your gkrellmrc file. The keyword is \"seti\" (take a look at the theme doc)."
};
static gchar	     seti_about_text[] =
"SETI@home GKrellM plugin version " MAJOR_VERSION "." MINOR_VERSION "." RELEASE_NUMBER "\n\n" \
"Copyright (C) 2000 Xavier Serpaggi\n" \
"xavier.serpaggi@libertysurf.fr\n" \
"http://perso.libertysurf.fr/serpaggi/\n\n" \
"SETI@home GKrellM plugin comes with ABSOLUTELY NO WARRANTY.\n" \
"This is free software, and you are welcome to redistribute\n" \
"it under certain conditions.\n"
"See the GNU Public License for details" ;


							 
/* check is there's a process with pid number equal to */
/* the one found int pid.sah file of seti client.      */
gint
seti_is_running(gint *pid)
{
	FILE	*f ;
	gchar 	 pid_file[255],
	 	 pid_string[10] ; /* A large pid, but who knows 8) */


	if ( !(f = fopen(state_file, "r")) ) {
		*pid = -1 ;
		return 0 ;
	}
	fclose(f) ;

	sprintf(pid_file, "%s/pid.sah", seti_file_path) ;
	if ( !(f = fopen(pid_file, "r")) ) {
		*pid = -1 ;
		return 0 ;
	}
	fgets(pid_string, 254, f) ;
	fclose (f) ;
	
	if ( sscanf(pid_string, "%d", pid) == 0 ) {
		*pid = -1 ;
		return 0 ;
	}

	sprintf(pid_file, "/proc/%d/status", *pid) ;
	if ( !(f = fopen(pid_file, "r")) )
		return 0 ;
	fclose(f) ;

	return 1 ;
}



static void
update_seti()
{
	Krell		*k ;
	FILE		*f ;
	static gdouble	 lprogress = 0 ;
	static gint	 tic_toc = 0 ;
	static gint	 previous_state = -1 ;
	static gint	 wu_nb = 0 ;
	gint		 state = 0 ;
	gint		 pid ;
	gchar		 buf[160] ;
	
	
	if ( seti_first_create || (state = seti_is_running(&pid)) ) {

		tic_toc = 1 ;

		if ( seti_first_create || GK.minute_tick ) {

			if (seti_first_create) {
				seti_first_create = 0 ;
				
				/* Get the current locale, as it seems sscanf wants it */
				locale_conv = localeconv() ;

				if (GK.debug) {
					printf("Let me introduce myself; I'm the SETI@home plugin for GKrellM.\n") ;
					if ( locale_conv ) printf("On this system the decimal point is [%s]\n", locale_conv->decimal_point) ;
				}
			}
			
			if ( (f = fopen(user_file, "r")) != NULL ) {
				while ( (fgets(buf, sizeof(buf), f)) != NULL ) {
					if ( strncmp(buf, "nresults", 8) == 0 ) {
						sscanf(buf, "nresults=%d", &wu_nb) ;

						if (GK.debug) 
							printf("SETI@home plugin progress: %g\n", lprogress);
	
					}
				}
				fclose(f) ;
			} else {
				wu_nb = 0 ;
			}
			
			if ( (f = fopen(state_file, "r")) != NULL ) {
	
				while ( (fgets(buf, sizeof(buf), f)) != NULL ) {
	
					if ( strncmp(buf, "prog", 4) == 0) {
	
						/*
						 | Take locale into account.
						 | I guess this is enough to have it work!
						 | But what happens if the scanf family does not use it?
						 */
						if ( locale_conv && locale_conv->decimal_point )
							buf[6] = locale_conv->decimal_point[0] ;

						sscanf(buf,"prog=%lg", &lprogress);
	
						if (GK.debug) 
							printf("SETI@home plugin progress: %g\n", lprogress);
	
						lprogress *= 100. ;
					}
				}
				fclose(f);
			}
	
			if ( GK.debug )
				printf("SETI@home plugin progess: %f\n", lprogress) ;
	
			k = KRELL(panel);
	
			/* GKrellM updates routines are designed to record differences between
			 |  the current data value and the data value from the previous call.
			 |  This is for handling continuously increasing data values.  However,
			 |  if you want to measure values between a fixed range, you can do so
			 |  by setting previous to zero before each update call.
			 */
			k->previous = 0;
			k->full_scale = 100;
			gkrellm_update_krell(panel, k, lprogress );
	
			/* Updating a krell (or drawing a decal) draws only into local
			 |  pixmaps.  After all krells and decals are "drawn", they must be
			 |  really drawn to the screen with gkrellm_draw_layers().
			 */
			gkrellm_draw_layers(panel);

			g_free(tooltip_text) ;
			tooltip_text = g_strdup_printf("Workunit %d - %.2f%% complete", wu_nb, lprogress) ;
			gtk_tooltips_set_tip( tooltip, panel->drawing_area, tooltip_text , NULL) ;
		}


		if ( previous_state != state ) {

			previous_state = state ;

			k = KRELL(panel);
	
			/* GKrellM updates routines are designed to record differences between
			 |  the current data value and the data value from the previous call.
			 |  This is for handling continuously increasing data values.  However,
			 |  if you want to measure values between a fixed range, you can do so
			 |  by setting previous to zero before each update call.
			 */
			k->previous = 0;
			k->full_scale = 100;
			gkrellm_update_krell(panel, k, lprogress );
	
			/* Updating a krell (or drawing a decal) draws only into local
			 |  pixmaps.  After all krells and decals are "drawn", they must be
			 |  really drawn to the screen with gkrellm_draw_layers().
			 */
			gkrellm_draw_layers(panel);

			g_free(tooltip_text) ;
			tooltip_text = g_strdup_printf("Workunit %d - %.2f%% complete", wu_nb, lprogress) ;
			gtk_tooltips_set_tip( tooltip, panel->drawing_area, tooltip_text , NULL) ;
		}


	} else  {       /* setiathome is not running */
		
		if ( GK.second_tick ) {
		
			/* I guess this is not a really good solution to make the */
			/* flying saucer blink. Give me a better one please.      */
			tic_toc = (tic_toc == 0) ;      
		
			k = KRELL(panel);
			k->previous = 0;
			k->full_scale = 100;
		
			gkrellm_update_krell(panel, k, (gdouble)tic_toc * lprogress) ;
			gkrellm_draw_layers(panel);
		}

		if ( previous_state != state ) {
			
			previous_state = state ;

			g_free(tooltip_text) ;
			tooltip_text = g_strdup_printf("Workunit %d - %.2f%% complete (idle)", wu_nb, lprogress) ;
			gtk_tooltips_set_tip( tooltip, panel->drawing_area, tooltip_text , NULL) ;
		}
	}
}


static gint
panel_expose_event(GtkWidget *widget, GdkEventExpose *ev)
{
	gdk_draw_pixmap(widget->window,
			widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
			panel->pixmap, ev->area.x, ev->area.y, ev->area.x, ev->area.y,
			ev->area.width, ev->area.height);
	return FALSE;
}



/* User has clicked panel. We should try to start/stop seti@home software */
/* client_stop() and client_start() inspired by e-setimon from Ben Wilson */
static void
client_stop (gint pid)
{
	int res ;
	/*
	char	runthis[256];


	sprintf (runthis, "kill %d\n", pid);

	system (runthis);
	*/

	/* This solution avoid to do a `system' call, and
	 | thus is less exposed to security issues.
	 */
	res = kill(pid, SIGTERM) ;

	if ( (res == -1) && (GK.debug) ) {
		printf("SETI@home plugin: unable to kill setiathome client pid %d\n", pid) ;
	}

	return;
}


static void
client_start(void)
{
	char runthis[256];
	
	sprintf (runthis, "cd %s ; %s/setiathome %s & \n", seti_file_path, seti_client_path, seti_args); 
   
	if (GK.debug)
		printf("SETI@home plugin: command passed to client = \"%s\"\n", runthis) ;

	  
	system (runthis);


	return;
}


static gint
panel_click_event(GtkWidget *widget, GdkEventButton *ev)
{
	gint pid ;

	if (ev->button == 1) {


		if (GK.debug)
			printf("Left click on SETI@home plugin: will try to start/stop client\n");


		if (seti_is_running(&pid))
			client_stop(pid) ;
		else if ( pid != -1 )
			client_start() ;
	} 
	   
		 
	return 1 ;
}


static void
create_seti(GtkWidget *vbox, gint first_create)
{
	Krell	   *k ;
	Style	   *style ;


	/* This create_plugin() routine is a create event routine which
	 |  will be called at every theme or size change.  The idea is to allocate
	 |  data structures and do some initialization only once (at first_create)
	 |  and do decal and krell creation and plugin local image loading each
	 |  time so theme changes will be picked up.
	 |  But you must be careful to destroy some things to avoid memory
	 |  leaks and crashes.  Many gkrellm functions will destroy images and
	 |  pixmaps for you automatically, but if you make any krells or decals,
	 |  you must destroy the krell and decal lists.  Here we make no decals.
	 */
	if (first_create)
		panel = gkrellm_panel_new0();
	else
		gkrellm_destroy_krell_list(panel);

		
	/* Create a krell.  A Krell structure is allocated and linked into
	 |  the list of krells pointed to by panel->krell.  By using the style_id
	 |  determined in init_plugin() we are asking for any custom theme images
	 |  for this specific plugin.
	 */
	style = gkrellm_meter_style(style_id);
	style->krell_x_hot = 24 ;
	style->krell_yoff = 3 ;
	style->krell_depth = 22 ;


	gkrellm_load_image("krell", krell_xpm, &krell_image, STYLE_NAME) ;
	gkrellm_load_image("bg_meter", bg_meter_xpm, &bg_meter_image, STYLE_NAME) ;
	k = gkrellm_create_krell(panel, krell_image, style);
	k->full_scale = 30;


	/* Configure panel calculates the panel height needed for the "Plugin" label.
	 |  and the krell.
	 */
	panel->textstyle = gkrellm_meter_textstyle(style_id);
	gkrellm_configure_panel(panel, "SETI", style);


	/* Build the configured panel with a background image and pack it into
	 |  the vbox assigned to this monitor.  The background image will be
	 |  the theme default or plugin specific if created by the theme author.
	 */
	gkrellm_create_panel(vbox, panel, bg_meter_image);
	gkrellm_monitor_height_adjust(panel->h);


	if ( tooltip == NULL ) {
		
		tooltip = gtk_tooltips_new() ;
		tooltip_text = g_strdup("SETI@home") ;
		
		gtk_tooltips_set_tip( tooltip, panel->drawing_area, tooltip_text , NULL) ;
		gtk_tooltips_set_delay(tooltip, 500);
	}


	if (first_create) {
		
		gtk_signal_connect(GTK_OBJECT (panel->drawing_area), "expose_event",
				(GtkSignalFunc) panel_expose_event, NULL);
		
		gtk_signal_connect(GTK_OBJECT (panel->drawing_area), "button_release_event",
				(GtkSignalFunc) panel_click_event, NULL);

		update_seti() ;
	}
}


static void
create_seti_config(GtkWidget *tab_vbox)
{
	GtkWidget       *frame;
	GtkWidget       *vbox;
	GtkWidget       *tabs;
	GtkWidget       *text;
	GtkWidget       *label;
	GtkWidget       *info_label;
	GtkWidget       *scrolled;


	tabs = gtk_notebook_new() ;
	gtk_notebook_set_tab_pos(GTK_NOTEBOOK(tabs), GTK_POS_TOP) ;
	gtk_box_pack_start(GTK_BOX(tab_vbox), tabs, TRUE, TRUE, 0) ;


	/* File location tab */
	frame = gtk_frame_new(NULL) ;
	gtk_container_border_width(GTK_CONTAINER(frame), 3) ;


	vbox = gtk_vbox_new(FALSE, 0) ;
	gtk_container_border_width(GTK_CONTAINER(vbox), 3) ;


	label = gtk_label_new("Path to state.sah file:") ;
	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0) ;
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT) ;
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0) ;


	seti_path_entry = gtk_entry_new_with_max_length(255) ;
	gtk_entry_set_text(GTK_ENTRY(seti_path_entry), seti_file_path) ;
	gtk_entry_set_editable(GTK_ENTRY(seti_path_entry), TRUE) ;
	gtk_box_pack_start(GTK_BOX(vbox), seti_path_entry, FALSE, FALSE, 0) ;



	/* avflinsch@att.net -- fix for client path */   
	label = gtk_label_new("Path to seti client:");
	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0) ;
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT) ;
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0) ;

	seti_client_entry = gtk_entry_new_with_max_length(255) ;
	gtk_entry_set_text(GTK_ENTRY(seti_client_entry), seti_client_path) ;
	gtk_entry_set_editable(GTK_ENTRY(seti_client_entry), TRUE) ;
	gtk_box_pack_start(GTK_BOX(vbox), seti_client_entry, FALSE, FALSE, 0) ;
   
	label = gtk_label_new("Arguments to the client:") ;
	gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0) ;
	gtk_label_set_justify(GTK_LABEL(label), GTK_JUSTIFY_LEFT) ;
	gtk_misc_set_alignment(GTK_MISC(label), 0, 0) ;


	seti_args_entry = gtk_entry_new_with_max_length(255);
	gtk_entry_set_text(GTK_ENTRY(seti_args_entry), seti_args) ;
	gtk_entry_set_editable(GTK_ENTRY(seti_args_entry), TRUE) ;
	gtk_box_pack_start(GTK_BOX(vbox), seti_args_entry, FALSE, FALSE, 0) ;


	label = gtk_label_new("Configs") ;
	gtk_container_add(GTK_CONTAINER(frame), vbox) ;
	gtk_notebook_append_page(GTK_NOTEBOOK(tabs), frame, label) ;
	
	/* Info tab */
	frame = gtk_frame_new(NULL) ;
	gtk_container_border_width(GTK_CONTAINER(frame), 3) ;


	scrolled = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_container_add(GTK_CONTAINER(frame), scrolled) ;
	
	text = gtk_text_new(NULL, NULL);
	
	gkrellm_add_info_text(text, seti_info_text, sizeof(seti_info_text) / sizeof(gchar *));
	gtk_text_set_editable(GTK_TEXT(text), FALSE);
	gtk_container_add(GTK_CONTAINER(scrolled), text);


	label = gtk_label_new("Info");
	gtk_notebook_append_page(GTK_NOTEBOOK(tabs), frame, label);


	/* About tab */
	info_label = gtk_label_new(seti_about_text);
	label = gtk_label_new("About");
	gtk_notebook_append_page(GTK_NOTEBOOK(tabs), info_label, label);
}


/*
 *      Saving config
 */


static void
save_seti_config(FILE *f)
{
	fprintf(f, "seti filepath %s\n", seti_file_path);
	fprintf(f, "seti args %s\n", seti_args);
	
	/* avflinsch@att.net -- fix for client path */
	fprintf(f, "seti clientpath %s\n", seti_client_path);
	   
	if (GK.debug)
		printf ("SETI@home plugin save:\n\t%s\n\t%s\n",
				seti_file_path,
				seti_args) ;
}


/*
 *  Loading config
 */


static void
load_seti_config(gchar *arg)
{
	gchar   seti_config[64], item[256];
	gint    n;


	n = sscanf(arg, "%s %[^\n]", seti_config, item);


	if (n == 2)
	{
		if (GK.debug)
			printf("SETI@home plugin config: %s item=\"%s\"\n",
					seti_config,
					item) ;


		if (strcmp(seti_config, "filepath") == 0) {
			
			g_free(seti_file_path) ;
			seti_file_path = g_strdup(item) ;
			
			g_free(state_file) ;
			state_file = g_strdup_printf("%s/state.sah", seti_file_path) ;
			
			g_free(user_file) ;
			user_file = g_strdup_printf("%s/user_info.sah", seti_file_path) ;
			
			if (GK.debug)
				printf ("SETI@home plugin load:\n\t%s\n\t%s\n\t%s\n",
						seti_file_path,
						state_file,
						user_file) ;


		} else if (strcmp(seti_config, "args") == 0) {
			
			g_free(seti_args) ;
			seti_args = g_strdup(item) ;
			
			if (GK.debug)
				printf ("SETI@home plugin load:\n\t%s\n",
						seti_args) ;
		   
		  /* avflinsch@att.net -- fix client path */  
		} else if (strcmp(seti_config, "clientpath") == 0) {
		       
		       g_free(seti_client_path) ;
		       seti_client_path = g_strdup(item) ; 
		       
		       if (GK.debug)
			       printf ("SETI@home plugin load:\n\t%s\n", 
					       seti_client_path) ;
		  
		 }
	}
}


/*
 *      Applying config.
 *      This includes only updating seti_file_path variable.
 */


static void
apply_seti_config()
{
	g_free(seti_file_path) ;
	seti_file_path = g_strdup(gtk_entry_get_text(GTK_ENTRY(seti_path_entry))) ;

	/* avflinsch@att.net -- fix client path */
	g_free(seti_client_path) ;
	seti_client_path = g_strdup(gtk_entry_get_text(GTK_ENTRY(seti_client_entry))) ;


   
	g_free(state_file) ;
	state_file = g_strdup_printf("%s/state.sah", seti_file_path) ;


	g_free(user_file) ;
	user_file = g_strdup_printf("%s/user_info.sah", seti_file_path) ;


	g_free(seti_args) ;
	seti_args = g_strdup(gtk_entry_get_text(GTK_ENTRY(seti_args_entry))) ;


	if (GK.debug)
		printf("SETI@home plugin apply:\n\t%s\n\t%s\n\t%s\n\t%s\n\t\%s\n",
				seti_file_path,
				seti_client_path,
				state_file,
				user_file,
				seti_args) ;
}



/* The monitor structure tells GKrellM how to call the plugin routines.
 */
static Monitor  plugin_mon      =
{
	CONFIG_NAME,	    /* Title for config clist.   */
	0,		      /* Id,  0 if a plugin       */
	create_seti,	    /* The create function      */
	update_seti,	    /* The update function      */


	create_seti_config,     /* The config tab create function   */
	apply_seti_config,      /* Apply the config function	*/


	save_seti_config,       /* Save user config		     */
	load_seti_config,       /* Load user config		     */
	"seti",		 /* config keyword		       */


	NULL,		   /* Undefined 2  */
	NULL,		   /* Undefined 1  */
	NULL,		   /* private      */


	SETI_PLACE,		/* Insert plugin before this monitor		    */


	NULL,		   /* Handle if a plugin, filled in by GKrellM     */
	NULL		    /* path if a plugin, filled in by GKrellM       */
};


/* All GKrellM plugins must have one global routine named init_plugin()
 |  which returns a pointer to a filled in monitor structure.
 */
Monitor *
init_plugin()
{
	/* If this next call is made, the background and krell images for this
	 |  plugin can be custom themed by putting bg_meter.png or krell.png in the
	 |  subdirectory STYLE_NAME of the theme directory.  Text colors (and
	 |  other things) can also be specified for the plugin with gkrellmrc
	 |  lines like:  StyleMeter  STYLE_NAME.textcolor orange black shadow
	 |  If no custom theming has been done, then all above calls using
	 |  style_id will be equivalent to style_id = DEFAULT_STYLE_ID.
	 */
	style_id = gkrellm_add_meter_style(&plugin_mon, STYLE_NAME);


	g_free(seti_file_path) ;
	seti_file_path = g_strdup("/opt/setiathome/") ;
	
	/* avflinsch@att.net -- fix client path */
	g_free(seti_client_path) ;
	seti_client_path = g_strdup("/opt/setiathome/") ;
   
	g_free(state_file) ;
	state_file = g_strdup_printf("%s/state.sah", seti_file_path) ;
	
	g_free(user_file) ;
	user_file = g_strdup_printf("%s/user_info.sah", seti_file_path) ;
	
	g_free(seti_args) ;
	seti_args = g_strdup("-nice 15 -email > /dev/null 2> /dev/null") ;


	tooltip = NULL ;

	return &plugin_mon;
}
